@@ -27,13 +27,18 @@ Typical usage example:
 """
 
 import json
+import platform
+import subprocess
+from pathlib import Path
 from typing import Any, Callable, Dict, List, Optional
 
 from local_operator.agents import AgentData, AgentEditFields, AgentRegistry
 from local_operator.config import Config, ConfigManager
 from local_operator.executor import LocalCodeExecutor
+from local_operator.notebook import save_code_history_to_notebook
 from local_operator.operator import ConversationRole
 from local_operator.tools import ToolRegistry
+from local_operator.types import AgentState
 
 
 def create_agent_from_conversation_tool(
@@ -72,26 +77,52 @@ def create_agent_from_conversation_tool(
             AgentData: The newly created agent's data
         """
         # Find index of last user message by iterating backwards
-        history = executor.conversation_history
+        conversation_history = executor.agent_state.conversation
+        execution_history = executor.agent_state.execution_history
         last_user_idx = None
 
-        for i in range(len(history) - 1, -1, -1):
-            if history[i]["role"].upper() == "USER":
+        for i in range(len(conversation_history) - 1, -1, -1):
+            if conversation_history[i].role == ConversationRole.USER:
                 last_user_idx = i
                 break
 
         # Save history up to the last user message (excluding the create
         # agent request itself)
         cutoff_idx = last_user_idx if last_user_idx is not None else 0
-        history_to_save = history[:cutoff_idx]
+        conversation_history_to_save = conversation_history[:cutoff_idx]
+        execution_history_to_save = execution_history[:cutoff_idx]
 
         new_agent = agent_registry.create_agent(
             AgentEditFields(
                 name=name,
                 security_prompt="",
+                hosting="",
+                model="",
+                description="",
+                last_message="",
+                temperature=None,
+                top_p=None,
+                top_k=None,
+                max_tokens=None,
+                stop=None,
+                frequency_penalty=None,
+                presence_penalty=None,
+                seed=None,
+                current_working_directory=None,
             )
         )
-        agent_registry.save_agent_conversation(new_agent.id, history_to_save)
+        agent_registry.save_agent_state(
+            new_agent.id,
+            AgentState(
+                version=new_agent.version,
+                conversation=conversation_history_to_save,
+                execution_history=execution_history_to_save,
+                learnings=executor.agent_state.learnings,
+                current_plan=executor.agent_state.current_plan,
+                instruction_details=executor.agent_state.instruction_details,
+                agent_system_prompt=executor.agent_state.agent_system_prompt,
+            ),
+        )
         return new_agent
 
     return create_agent_from_conversation
@@ -141,19 +172,30 @@ def save_agent_training_tool(
             raise ValueError("No current agent set in executor")
 
         # Find index of last user message by iterating backwards
-        history = executor.conversation_history
+        conversation_history = executor.agent_state.conversation
+        execution_history = executor.agent_state.execution_history
         last_user_idx = None
 
-        for i in range(len(history) - 1, -1, -1):
-            if history[i]["role"] == ConversationRole.USER.value:
+        for i in range(len(conversation_history) - 1, -1, -1):
+            if conversation_history[i].role == ConversationRole.USER:
                 last_user_idx = i
                 break
 
         # Save history up to the last user message (excluding the save request itself)
         cutoff_idx = last_user_idx if last_user_idx is not None else 0
-        history_to_save = history[:cutoff_idx]
+        conversation_history_to_save = conversation_history[:cutoff_idx]
+        execution_history_to_save = execution_history[:cutoff_idx]
 
-        agent_registry.save_agent_conversation(executor.agent.id, history_to_save)
+        agent_state_to_save = AgentState(
+            conversation=conversation_history_to_save,
+            execution_history=execution_history_to_save,
+            **executor.agent_state.model_dump(exclude={"conversation", "execution_history"}),
+        )
+
+        agent_registry.save_agent_state(
+            executor.agent.id,
+            agent_state_to_save,
+        )
         return executor.agent
 
     return save_agent_training
@@ -232,6 +274,19 @@ def create_agent_tool(agent_registry: AgentRegistry) -> Callable[[str, Optional[
             AgentEditFields(
                 name=name,
                 security_prompt=security_prompt or "",
+                hosting="",
+                model="",
+                description="",
+                last_message="",
+                temperature=None,
+                top_p=None,
+                top_k=None,
+                max_tokens=None,
+                stop=None,
+                frequency_penalty=None,
+                presence_penalty=None,
+                seed=None,
+                current_working_directory=None,
             )
         )
 
@@ -338,7 +393,7 @@ def get_agent_info_tool(
     return get_agent_info
 
 
-def save_conversation_tool(
+def save_conversation_raw_json_tool(
     executor: LocalCodeExecutor,
 ) -> Callable[[str], None]:
     """Create a tool function that saves conversation history to disk in JSON format.
@@ -359,8 +414,8 @@ def save_conversation_tool(
         RuntimeError: If there are unexpected issues during the save operation
     """
 
-    def save_conversation(filename: str) -> None:
-        """Save the current Local Operator conversation history to a JSON file.
+    def save_conversation_raw_json(filename: str) -> None:
+        """Save the current Local Operator conversation history to a raw JSON file.
 
         Exports the complete conversation history including all messages between user and agent,
         commands executed, and their results. The file can be used for analysis or to initialize
@@ -383,7 +438,7 @@ def save_conversation_tool(
             # Save to JSON file
             with open(filename, "w", encoding="utf-8") as f:
                 json.dump(
-                    [msg if isinstance(msg, dict) else msg.dict() for msg in conversation],
+                    [msg.dict() for msg in conversation],
                     f,
                     indent=2,
                     ensure_ascii=False,
@@ -394,7 +449,7 @@ def save_conversation_tool(
         except Exception as e:
             raise RuntimeError(f"Error saving conversation: {str(e)}")
 
-    return save_conversation
+    return save_conversation_raw_json
 
 
 def get_config_tool(config_manager: ConfigManager) -> Callable[[], Config]:
@@ -470,6 +525,137 @@ def update_config_tool(config_manager: ConfigManager) -> Callable[[Dict[str, Any
     return update_config
 
 
+def open_agents_config_tool(agent_registry: AgentRegistry) -> Callable[[], None]:
+    """Create a tool function that opens the agents directory.
+
+    This function returns a callable that opens the agents directory in the default system file
+    explorer. The directory location is determined from the agent registry's config directory.
+
+    Args:
+        agent_registry: The AgentRegistry instance containing the config directory path
+
+    Returns:
+        Callable[[], None]: A function that opens the agents directory
+
+    Raises:
+        RuntimeError: If there are issues opening the directory
+    """
+
+    def open_agents_config() -> None:
+        """Open the agents directory in the default system file explorer.
+
+        Opens the agents directory located in the agent registry's config directory
+        using the system's default file explorer.
+
+        Raises:
+            RuntimeError: If the directory cannot be opened
+        """
+        try:
+            agents_dir = agent_registry.agents_dir
+            if not agents_dir.exists():
+                raise RuntimeError(f"Agents directory not found at {agents_dir}")
+
+            system = platform.system()
+            if system == "Darwin":  # macOS
+                subprocess.run(["open", str(agents_dir)], check=True)
+            elif system == "Windows":
+                subprocess.run(["start", str(agents_dir)], shell=True, check=True)
+            else:  # Linux and other Unix-like
+                subprocess.run(["xdg-open", str(agents_dir)], check=True)
+
+        except Exception as e:
+            raise RuntimeError(f"Failed to open agents directory: {str(e)}")
+
+    return open_agents_config
+
+
+def open_settings_config_tool(config_manager: ConfigManager) -> Callable[[], None]:
+    """Create a tool function that opens the main settings configuration file.
+
+    This function returns a callable that opens the config.yml file in the default system editor.
+    The file location is determined from the config manager's config directory.
+
+    Args:
+        config_manager: The ConfigManager instance containing the config directory path
+
+    Returns:
+        Callable[[], None]: A function that opens the settings configuration file
+
+    Raises:
+        RuntimeError: If there are issues opening the configuration file
+    """
+
+    def open_settings_config() -> None:
+        """Open the settings configuration file in the default system editor.
+
+        Opens the config.yml file located in the config manager's directory
+        using the system's default application for YAML files.
+
+        Raises:
+            RuntimeError: If the configuration file cannot be opened
+        """
+        try:
+            config_file = config_manager.config_dir / "config.yml"
+            if not config_file.exists():
+                raise RuntimeError(f"Settings configuration file not found at {config_file}")
+
+            system = platform.system()
+            if system == "Darwin":  # macOS
+                subprocess.run(["open", str(config_file)], check=True)
+            elif system == "Windows":
+                subprocess.run(["start", str(config_file)], shell=True, check=True)
+            else:  # Linux and other Unix-like
+                subprocess.run(["xdg-open", str(config_file)], check=True)
+
+        except Exception as e:
+            raise RuntimeError(f"Failed to open settings configuration file: {str(e)}")
+
+    return open_settings_config
+
+
+def save_conversation_history_to_notebook_tool(
+    executor: LocalCodeExecutor,
+) -> Callable[[str], None]:
+    """Create a tool function that saves conversation history to a notebook in ipynb format.
+
+    This function returns a callable that, when called, saves the executed code blocks
+    along with their outputs into an IPython notebook file (.ipynb). The notebook
+    is saved in JSON format.
+
+    Returns:
+        Callable[[str], None]: A function that saves the code blocks to a notebook file.
+    """
+
+    def save_conversation_history_to_notebook(file_path: str = "notebook.ipynb") -> None:
+        """Save the conversation history to an IPython notebook file (.ipynb).
+
+        This function retrieves the code blocks and their execution results from the
+        executor, formats them as notebook cells, and saves them to a .ipynb file
+        in JSON format.
+
+        Args:
+            file_path (str): The path to save the notebook to. Defaults to "notebook.ipynb".
+
+        Raises:
+            Exception: If there is an error during notebook creation or file saving.
+        """
+        try:
+            save_code_history_to_notebook(
+                code_history=executor.agent_state.execution_history,
+                model_configuration=executor.model_configuration,
+                max_conversation_history=executor.max_conversation_history,
+                detail_conversation_length=executor.detail_conversation_length,
+                max_learnings_history=executor.max_learnings_history,
+                file_path=Path(file_path),
+            )
+            print(f"Notebook saved to {file_path}")
+
+        except Exception as e:
+            raise Exception(f"Failed to save conversation history to notebook: {str(e)}")
+
+    return save_conversation_history_to_notebook
+
+
 def add_admin_tools(
     tool_registry: ToolRegistry,
     executor: LocalCodeExecutor,
@@ -485,10 +671,12 @@ def add_admin_tools(
     - edit_agent_tool
     - delete_agent_tool
     - get_agent_info_tool
-    - save_conversation_tool
+    - save_conversation_raw_json_tool
     - get_config_tool
     - update_config_tool
-
+    - open_agents_config_tool
+    - open_settings_config_tool
+    - save_conversation_history_to_notebook_tool
     Args:
         tool_registry: The ToolRegistry instance to add tools to
     """
@@ -512,8 +700,8 @@ def add_admin_tools(
         get_agent_info_tool(agent_registry),
     )
     tool_registry.add_tool(
-        "save_conversation",
-        save_conversation_tool(executor),
+        "save_conversation_raw_json",
+        save_conversation_raw_json_tool(executor),
     )
     tool_registry.add_tool(
         "get_config",
@@ -527,3 +715,15 @@ def add_admin_tools(
         "save_agent_training",
         save_agent_training_tool(executor, agent_registry),
     )
+    tool_registry.add_tool(
+        "open_agents_config",
+        open_agents_config_tool(agent_registry),
+    )
+    tool_registry.add_tool(
+        "open_settings_config",
+        open_settings_config_tool(config_manager),
+    )
+    tool_registry.add_tool(
+        "save_conversation_history_to_notebook",
+        save_conversation_history_to_notebook_tool(executor),
+    )
