@@ -18,20 +18,29 @@ import argparse
 import asyncio
 import math
 import os
+import platform
+import subprocess
 import traceback
 from importlib.metadata import version
 from pathlib import Path
+from typing import Optional
 
 import uvicorn
+from pydantic import SecretStr
 
 from local_operator.admin import add_admin_tools
 from local_operator.agents import AgentEditFields, AgentRegistry
+from local_operator.clients.fal import FalClient
+from local_operator.clients.openrouter import OpenRouterClient
+from local_operator.clients.serpapi import SerpApiClient
+from local_operator.clients.tavily import TavilyClient
 from local_operator.config import ConfigManager
 from local_operator.credentials import CredentialManager
 from local_operator.executor import LocalCodeExecutor
-from local_operator.model import configure_model
+from local_operator.model.configure import configure_model, validate_model
 from local_operator.operator import Operator, OperatorType
 from local_operator.tools import ToolRegistry
+from local_operator.types import AgentState
 
 CLI_DESCRIPTION = """
     Local Operator - An environment for agentic AI models to perform tasks on the local device.
@@ -110,28 +119,72 @@ def build_cli_parser() -> argparse.ArgumentParser:
         "claude-3-5-sonnet-20240620, moonshot-v1-32k, qwen-plus, gemini-2.0-flash, "
         "mistral-large-latest, test-model, deepseek/deepseek-chat)",
     )
+    parser.add_argument(
+        "--run-in",
+        type=str,
+        help="The working directory to run the operator in.  Must be a valid directory.",
+        dest="run_in",
+    )
     subparsers = parser.add_subparsers(dest="subcommand")
-
     # Credential command
     credential_parser = subparsers.add_parser(
         "credential",
         help="Manage API keys and credentials for different hosting platforms",
         parents=[parent_parser],
     )
-    credential_parser.add_argument(
-        "--key",
-        type=str,
-        required=True,
-        help="Credential key to update (e.g., DEEPSEEK_API_KEY, "
-        "OPENAI_API_KEY, ANTHROPIC_API_KEY, KIMI_API_KEY, ALIBABA_CLOUD_API_KEY, "
-        "GOOGLE_AI_STUDIO_API_KEY, MISTRAL_API_KEY, OPENROUTER_API_KEY)",
+    credential_subparsers = credential_parser.add_subparsers(dest="credential_command")
+    credential_update_parser = credential_subparsers.add_parser(
+        "update", help="Update a credential", parents=[parent_parser]
     )
 
+    credential_delete_parser = credential_subparsers.add_parser(
+        "delete", help="Delete a credential", parents=[parent_parser]
+    )
+
+    credential_key_help = (
+        "Credential key to manage (e.g., DEEPSEEK_API_KEY, OPENAI_API_KEY, "
+        "ANTHROPIC_API_KEY, KIMI_API_KEY, ALIBABA_CLOUD_API_KEY, GOOGLE_AI_STUDIO_API_KEY, "
+        "MISTRAL_API_KEY, OPENROUTER_API_KEY)"
+    )
+
+    credential_update_parser.add_argument("key", type=str, help=credential_key_help)
+    credential_delete_parser.add_argument("key", type=str, help=credential_key_help)
+
     # Config command
     config_parser = subparsers.add_parser(
         "config", help="Manage configuration settings", parents=[parent_parser]
     )
     config_subparsers = config_parser.add_subparsers(dest="config_command")
+    # Open command
+    config_subparsers.add_parser(
+        "open", help="Open the configuration file in the default editor", parents=[parent_parser]
+    )
+    # Edit command
+    config_edit_parser = config_subparsers.add_parser(
+        "edit",
+        help="Edit a specific configuration value in the config file",
+        parents=[parent_parser],
+    )
+    config_edit_parser.add_argument(
+        "key",
+        type=str,
+        help="Configuration key to update (e.g., hosting, model_name, conversation_length, "
+        "detail_length, max_learnings_history, auto_save_conversation)",
+    )
+    config_edit_parser.add_argument(
+        "value",
+        type=str,
+        help="New value for the configuration key (type is automatically converted "
+        "based on the key)",
+    )
+
+    # List command
+    config_subparsers.add_parser(
+        "list",
+        help="List available configuration options and their descriptions",
+        parents=[parent_parser],
+    )
+
     config_subparsers.add_parser(
         "create", help="Create a new configuration file", parents=[parent_parser]
     )
@@ -184,8 +237,8 @@ def build_cli_parser() -> argparse.ArgumentParser:
     serve_parser.add_argument(
         "--port",
         type=int,
-        default=8080,
-        help="Port for the server (default: 8080)",
+        default=1111,
+        help="Port for the server (default: 1111)",
     )
     serve_parser.add_argument(
         "--reload",
@@ -208,12 +261,18 @@ def build_cli_parser() -> argparse.ArgumentParser:
     return parser
 
 
-def credential_command(args: argparse.Namespace) -> int:
+def credential_update_command(args: argparse.Namespace) -> int:
     credential_manager = CredentialManager(Path.home() / ".local-operator")
     credential_manager.prompt_for_credential(args.key, reason="update requested")
     return 0
 
 
+def credential_delete_command(args: argparse.Namespace) -> int:
+    credential_manager = CredentialManager(Path.home() / ".local-operator")
+    credential_manager.set_credential(args.key, "")
+    return 0
+
+
 def config_create_command() -> int:
     """Create a new configuration file."""
     config_manager = ConfigManager(Path.home() / ".local-operator")
@@ -222,12 +281,100 @@ def config_create_command() -> int:
     return 0
 
 
+def config_open_command() -> int:
+    """Open the configuration file using the default system editor."""
+    config_path = Path.home() / ".local-operator" / "config.yml"
+    if not config_path.exists():
+        print(
+            "\n\033[1;31mError: Configuration file does not exist.  Create one with "
+            "`config create`.\033[0m"
+        )
+        return -1
+
+    try:
+        if platform.system() == "Windows":
+            subprocess.run(["start", str(config_path)], shell=True, check=True)
+        elif platform.system() == "Darwin":
+            subprocess.run(["open", str(config_path)], check=True)
+        else:
+            subprocess.run(["xdg-open", str(config_path)], check=True)
+        print(f"Opened configuration file at {config_path}")
+        return 0
+    except Exception as e:
+        print(f"\n\033[1;31mError opening configuration file: {e}\033[0m")
+        return -1
+
+
+def config_edit_command(args: argparse.Namespace) -> int:
+    """Edit a configuration value."""
+    config_manager = ConfigManager(Path.home() / ".local-operator")
+    try:
+        # Parse the value to the appropriate type
+        value = args.value
+        # Try to convert to int
+        try:
+            if value.isdigit() or (value.startswith("-") and value[1:].isdigit()):
+                value = int(value)
+            # Try to convert to float
+            elif value.replace(".", "", 1).isdigit() or (
+                value.startswith("-") and value[1:].replace(".", "", 1).isdigit()
+            ):
+                value = float(value)
+            # Try to convert to boolean
+            elif value.lower() in ("true", "false"):
+                value = value.lower() == "true"
+            # Handle null/None values
+            elif value.lower() in ("null", "none"):
+                value = None
+        except (ValueError, AttributeError):
+            # Keep as string if conversion fails
+            pass
+
+        config_manager.update_config(
+            {args.key: value},
+            write=True,
+        )
+
+        print(f"Successfully updated {args.key} to {value}")
+        return 0
+    except KeyError:
+        print(f"\n\033[1;31mError: Invalid configuration key: {args.key}\033[0m")
+        return -1
+    except Exception as e:
+        print(f"\n\033[1;31mError updating configuration: {e}\033[0m")
+        return -1
+
+
+def config_list_command() -> int:
+    """List available configuration options and their descriptions."""
+    config_manager = ConfigManager(Path.home() / ".local-operator")
+    config = config_manager.get_config()
+
+    # Configuration descriptions
+    descriptions = {
+        "hosting": "AI provider platform (e.g., openai, deepseek, anthropic, openrouter)",
+        "model_name": "The specific model to use for interactions",
+        "conversation_length": "Maximum number of messages to keep in conversation history",
+        "detail_length": "Number of recent messages to leave unsummarized in conversation history",
+        "max_learnings_history": "Maximum number of learning entries to retain",
+        "auto_save_conversation": "Whether to automatically save conversations",
+    }
+
+    print("\n\033[1;32m╭─ Configuration Options ───────────────────────\033[0m")
+    for key, value in config.values.items():
+        description = descriptions.get(key, "No description available")
+        print(f"\033[1;32m│ {key}: {value}\033[0m")
+        print(f"\033[1;32m│   Description: {description}\033[0m")
+    print("\033[1;32m╰──────────────────────────────────────────────\033[0m")
+    return 0
+
+
 def serve_command(host: str, port: int, reload: bool) -> int:
     """
     Start the FastAPI server using uvicorn.
     """
     print(f"Starting server at http://{host}:{port}")
-    uvicorn.run("local_operator.server:app", host=host, port=port, reload=reload)
+    uvicorn.run("local_operator.server.app:app", host=host, port=port, reload=reload)
     return 0
 
 
@@ -255,11 +402,13 @@ def agents_list_command(args: argparse.Namespace, agent_registry: AgentRegistry)
         is_last = i == len(page_agents) - 1
         branch = "└──" if is_last else "├──"
         print(f"\033[1;32m│ {branch} Agent {start_idx + i + 1}\033[0m")
-        left_bar = "│ │" if not is_last else "│ "
+        left_bar = "│ │" if not is_last else "│  "
         print(f"\033[1;32m{left_bar}   • Name: {agent.name}\033[0m")
         print(f"\033[1;32m{left_bar}   • ID: {agent.id}\033[0m")
         print(f"\033[1;32m{left_bar}   • Created: {agent.created_date}\033[0m")
         print(f"\033[1;32m{left_bar}   • Version: {agent.version}\033[0m")
+        print(f"\033[1;32m{left_bar}   • Hosting: {agent.hosting or "default"}\033[0m")
+        print(f"\033[1;32m{left_bar}   • Model: {agent.model or "default"}\033[0m")
         if not is_last:
             print("\033[1;32m│ │\033[0m")
 
@@ -286,7 +435,25 @@ def agents_create_command(name: str, agent_registry: AgentRegistry) -> int:
             print("\n\033[1;31mAgent creation cancelled\033[0m")
             return -1
 
-    agent = agent_registry.create_agent(AgentEditFields(name=name, security_prompt=None))
+    agent = agent_registry.create_agent(
+        AgentEditFields(
+            name=name,
+            security_prompt=None,
+            hosting=None,
+            model=None,
+            description=None,
+            last_message=None,
+            temperature=None,
+            top_p=None,
+            top_k=None,
+            max_tokens=None,
+            stop=None,
+            frequency_penalty=None,
+            presence_penalty=None,
+            seed=None,
+            current_working_directory=None,
+        )
+    )
     print("\n\033[1;32m╭─ Created New Agent ───────────────────────────\033[0m")
     print(f"\033[1;32m│ Name: {agent.name}\033[0m")
     print(f"\033[1;32m│ ID: {agent.id}\033[0m")
@@ -311,27 +478,48 @@ def agents_delete_command(name: str, agent_registry: AgentRegistry) -> int:
 
 
 def build_tool_registry(
-    executor: LocalCodeExecutor, agent_registry: AgentRegistry, config_manager: ConfigManager
+    executor: LocalCodeExecutor,
+    agent_registry: AgentRegistry,
+    config_manager: ConfigManager,
+    credential_manager: CredentialManager,
 ) -> ToolRegistry:
-    """Build and initialize the tool registry with agent management tools.
+    """Build and initialize the tool registry.
 
-    This function creates a new ToolRegistry instance and registers the core agent management tools:
-    - create_agent_from_conversation: Creates a new agent from the current conversation
-    - edit_agent: Modifies an existing agent's properties
-    - delete_agent: Removes an agent from the registry
-    - get_agent_info: Retrieves information about agents
+    This function creates a new ToolRegistry instance, initializes default tools,
+    and adds admin tools for agent management. It also sets up SERP API and Tavily
+    API clients if the corresponding API keys are available.
 
     Args:
-        executor: The LocalCodeExecutor instance containing conversation history
-        agent_registry: The AgentRegistry for managing agents
-        config_manager: The ConfigManager for managing configuration
+        executor (LocalCodeExecutor): The LocalCodeExecutor instance for conversation history.
+        agent_registry (AgentRegistry): The AgentRegistry for managing agents.
+        config_manager (ConfigManager): The ConfigManager for managing configuration.
+        credential_manager (CredentialManager): The CredentialManager for managing credentials.
 
     Returns:
-        ToolRegistry: The initialized tool registry with all agent management tools registered
+        ToolRegistry: The initialized tool registry with all tools registered.
     """
     tool_registry = ToolRegistry()
+
+    serp_api_key = credential_manager.get_credential("SERP_API_KEY")
+    tavily_api_key = credential_manager.get_credential("TAVILY_API_KEY")
+    fal_api_key = credential_manager.get_credential("FAL_API_KEY")
+
+    if serp_api_key:
+        serp_api_client = SerpApiClient(serp_api_key)
+        tool_registry.set_serp_api_client(serp_api_client)
+
+    if tavily_api_key:
+        tavily_client = TavilyClient(tavily_api_key)
+        tool_registry.set_tavily_client(tavily_client)
+
+    if fal_api_key:
+        fal_client = FalClient(fal_api_key)
+        tool_registry.set_fal_client(fal_client)
+
     tool_registry.init_tools()
+
     add_admin_tools(tool_registry, executor, agent_registry, config_manager)
+
     return tool_registry
 
 
@@ -344,12 +532,28 @@ def main() -> int:
 
         config_dir = Path.home() / ".local-operator"
         agents_dir = config_dir / "agents"
+        agent_home_dir = Path.home() / "local-operator-home"
+
+        # Create the agent home directory if it doesn't exist
+        if not agent_home_dir.exists():
+            agent_home_dir.mkdir(parents=True, exist_ok=True)
 
         if args.subcommand == "credential":
-            return credential_command(args)
+            if args.credential_command == "update":
+                return credential_update_command(args)
+            elif args.credential_command == "delete":
+                return credential_delete_command(args)
+            else:
+                parser.error(f"Invalid credential command: {args.credential_command}")
         elif args.subcommand == "config":
             if args.config_command == "create":
                 return config_create_command()
+            elif args.config_command == "open":
+                return config_open_command()
+            elif args.config_command == "edit":
+                return config_edit_command(args)
+            elif args.config_command == "list":
+                return config_list_command()
             else:
                 parser.error(f"Invalid config command: {args.config_command}")
         elif args.subcommand == "agents":
@@ -373,19 +577,14 @@ def main() -> int:
         # Override config with CLI args where provided
         config_manager.update_config_from_args(args)
 
-        # Get final config values
-        hosting = config_manager.get_config_value("hosting")
-        model = config_manager.get_config_value("model_name")
-
-        model_instance = configure_model(hosting, model, credential_manager)
-
-        if not model_instance:
-            error_msg = (
-                f"\n\033[1;31mError: Model not found for hosting: "
-                f"{hosting} and model: {model}\033[0m"
-            )
-            print(error_msg)
-            return -1
+        # Set working directory if provided and valid
+        if args.run_in:
+            run_in_path = Path(args.run_in).resolve()
+            if not run_in_path.is_dir():
+                print(f"\n\033[1;31mError: Invalid working directory: {args.run_in}\033[0m")
+                return -1
+            os.chdir(run_in_path)
+            print(f"\n\033[1;32mSetting working directory to: {run_in_path}\033[0m")
 
         # Get agent if name provided
         agent = None
@@ -397,7 +596,23 @@ def main() -> int:
                     f"Creating new agent...\033[0m"
                 )
                 agent = agent_registry.create_agent(
-                    AgentEditFields(name=args.agent_name, security_prompt=None)
+                    AgentEditFields(
+                        name=args.agent_name,
+                        security_prompt=None,
+                        hosting=None,
+                        model=None,
+                        description=None,
+                        last_message=None,
+                        temperature=None,
+                        top_p=None,
+                        top_k=None,
+                        max_tokens=None,
+                        stop=None,
+                        frequency_penalty=None,
+                        presence_penalty=None,
+                        seed=None,
+                        current_working_directory=None,
+                    )
                 )
                 print("\n\033[1;32m╭─ Created New Agent ───────────────────────────\033[0m")
                 print(f"\033[1;32m│ Name: {agent.name}\033[0m")
@@ -406,42 +621,119 @@ def main() -> int:
                 print(f"\033[1;32m│ Version: {agent.version}\033[0m")
                 print("\033[1;32m╰──────────────────────────────────────────────────\033[0m\n")
 
+        hosting = config_manager.get_config_value("hosting")
+        model_name = config_manager.get_config_value("model_name")
+
+        chat_args = {}
+
         if agent:
             # Get conversation history if agent name provided
-            conversation_history = agent_registry.load_agent_conversation(agent.id)
+            agent_state = agent_registry.load_agent_state(agent.id)
+
+            # Use agent's hosting and model if provided
+            if agent.hosting:
+                hosting = agent.hosting
+            if agent.model:
+                model_name = agent.model
+            if agent.temperature:
+                chat_args["temperature"] = agent.temperature
+            if agent.top_p:
+                chat_args["top_p"] = agent.top_p
+            if agent.top_k:
+                chat_args["top_k"] = agent.top_k
+            if agent.max_tokens:
+                chat_args["max_tokens"] = agent.max_tokens
+            if agent.stop:
+                chat_args["stop"] = agent.stop
+            if agent.frequency_penalty:
+                chat_args["frequency_penalty"] = agent.frequency_penalty
+            if agent.presence_penalty:
+                chat_args["presence_penalty"] = agent.presence_penalty
+            if agent.seed:
+                chat_args["seed"] = agent.seed
+
         else:
-            conversation_history = []
+            agent_state = AgentState(
+                version="",
+                conversation=[],
+                execution_history=[],
+                learnings=[],
+                current_plan=None,
+                instruction_details=None,
+                agent_system_prompt=None,
+            )
+
+        model_info_client: Optional[OpenRouterClient] = None
+
+        if hosting == "openrouter":
+            model_info_client = OpenRouterClient(
+                credential_manager.get_credential("OPENROUTER_API_KEY")
+            )
+
+        model_configuration = configure_model(
+            hosting, model_name, credential_manager, model_info_client, **chat_args
+        )
+
+        if not model_configuration.instance:
+            error_msg = (
+                f"\n\033[1;31mError: Model not found for hosting: "
+                f"{hosting} and model: {model_name}\033[0m"
+            )
+            print(error_msg)
+            return -1
+
+        validate_model(hosting, model_name, model_configuration.api_key or SecretStr(""))
 
         training_mode = False
         if args.train:
             training_mode = True
 
+        single_execution_mode = args.subcommand == "exec"
+
+        auto_save_conversation = config_manager.get_config_value("auto_save_conversation", False)
+
+        # If autosave is enabled, create an autosave agent if it doesn't exist already
+        if auto_save_conversation and not single_execution_mode:
+            agent_registry.create_autosave_agent()
+
         executor = LocalCodeExecutor(
-            model_instance,
-            detail_conversation_length=config_manager.get_config_value("detail_length", 10),
-            conversation_history=conversation_history,
+            model_configuration=model_configuration,
+            detail_conversation_length=config_manager.get_config_value("detail_length", 35),
+            max_conversation_history=config_manager.get_config_value(
+                "max_conversation_history", 100
+            ),
+            max_learnings_history=config_manager.get_config_value("max_learnings_history", 50),
             agent=agent,
+            agent_registry=agent_registry,
+            agent_state=agent_state,
+            persist_conversation=training_mode,
         )
 
         operator = Operator(
             executor=executor,
             credential_manager=credential_manager,
             config_manager=config_manager,
-            model_instance=model_instance,
+            model_configuration=model_configuration,
             type=OperatorType.CLI,
             agent_registry=agent_registry,
             current_agent=agent,
-            training_mode=training_mode,
+            auto_save_conversation=auto_save_conversation and not single_execution_mode,
+            persist_agent_conversation=auto_save_conversation and not single_execution_mode,
+        )
+
+        tool_registry = build_tool_registry(
+            executor, agent_registry, config_manager, credential_manager
         )
 
-        tool_registry = build_tool_registry(executor, agent_registry, config_manager)
         executor.set_tool_registry(tool_registry)
 
+        executor.load_agent_state(agent_state)
+
         # Start the async chat interface or execute single command
-        if args.subcommand == "exec":
-            message = asyncio.run(operator.execute_single_command(args.command))
-            if message:
-                print(message.response)
+        if single_execution_mode:
+            _, final_response = asyncio.run(operator.execute_single_command(args.command))
+            if final_response:
+                print(final_response)
             return 0
         else:
             asyncio.run(operator.chat())
